from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import DictSchema, optional


def test_dict_declaration():
    with when:
        sch = schema.dict

    with then:
        assert isinstance(sch, DictSchema)


def test_dict_empty_keys_declaration():
    with given:
        keys = {}

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == keys


def test_dict_invalid_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.dict([])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.dict` value must be an instance of 'dict', "
                                        "instance of 'list' [] given")


def test_dict_keys_declaration():
    with given:
        keys = {
            "id": schema.int(42),
            "name": schema.str("banana")
        }

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == {key: (val, False) for key, val in keys.items()}


def test_dict_optional_keys_declaration():
    with given:
        keys = {
            "id": schema.int(42),
            "name": schema.str("banana"),
            optional("created_at"): schema.int,
        }
        props = {
            "id": (schema.int(42), False),
            "name": (schema.str("banana"), False),
            "created_at": (schema.int, True),
        }

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == props


def test_dict_already_declared_declaration_error():
    with given:
        keys = {}

    with when, raises(Exception) as exception:
        schema.dict(keys)(keys)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.dict({})` is already declared"


def test_dict_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.dict({"key": {}})

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.dict` value must be an instance of 'Schema', "
                                        "instance of 'dict' {} given")



from typing import Any

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import make_already_declared_error
from ._schema import Schema

__all__ = ("ConstSchema", "ConstProps",)


class ConstProps(Props):
    @property
    def value(self) -> Nilable[Any]:
        return self.get("value")


class ConstSchema(Schema[ConstProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_const(self, **kwargs)

    def __call__(self, /, value: Any) -> "ConstSchema":
        if self.props.value is not Nil:
            raise make_already_declared_error(self)
        return self.__class__(self.props.update(value=value))



from baby_steps import given, then, when

from district42 import optional, represent, schema


def test_dict_representation():
    with given:
        sch = schema.dict

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.dict"


def test_dict_empty_representation():
    with given:
        sch = schema.dict({})

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.dict({})"


def test_dict_one_key_representation():
    with given:
        sch = schema.dict({"id": schema.int})

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int",
            "})"
        ])


def test_dict_many_keys_representation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str("banana")
        })

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int,",
            "    'name': schema.str('banana')",
            "})",
        ])


def test_dict_optional_key_representation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            optional("name"): schema.str,
        })

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int,",
            "    optional('name'): schema.str",
            "})"
        ])


def test_dict_nested_keys_representation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "user": schema.dict({
                "id": schema.int,
                "name": schema.str("banana")
            }),
            "is_deleted": schema.bool
        })

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int,",
            "    'user': schema.dict({",
            "        'id': schema.int,",
            "        'name': schema.str('banana')",
            "    }),",
            "    'is_deleted': schema.bool",
            "})",
        ])


def test_dict_relaxed_empty_representation():
    with given:
        sch = schema.dict({...: ...})

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.dict({...: ...})"


def test_dict_relaxed_one_key_representation():
    with given:
        sch = schema.dict({"id": schema.int, ...: ...})

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int,",
            "    ...: ...",
            "})",
        ])


def test_dict_relaxed_many_keys_representation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str("banana"),
            ...: ...,
        })

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.dict({",
            "    'id': schema.int,",
            "    'name': schema.str('banana'),",
            "    ...: ...",
            "})",
        ])



from typing import Any

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import (
    make_already_declared_error,
    make_incorrect_max_error,
    make_incorrect_min_error,
    make_invalid_type_error,
)
from ._schema import Schema

__all__ = ("FloatSchema", "FloatProps",)


class FloatProps(Props):
    @property
    def value(self) -> Nilable[float]:
        return self.get("value")

    @property
    def min(self) -> Nilable[float]:
        return self.get("min")

    @property
    def max(self) -> Nilable[float]:
        return self.get("max")


class FloatSchema(Schema[FloatProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_float(self, **kwargs)

    def __call__(self, /, value: float) -> "FloatSchema":
        if not isinstance(value, float):
            raise make_invalid_type_error(self, value, (float,))

        if self.props.value is not Nil:
            raise make_already_declared_error(self)

        if (self.props.min is not Nil) or (self.props.max is not Nil):
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(value=value))

    def min(self, /, value: float) -> "FloatSchema":
        if not isinstance(value, float):
            raise make_invalid_type_error(self, value, (float,))

        if self.props.min is not Nil:
            raise make_already_declared_error(self)

        if (self.props.value is not Nil) and (value > self.props.value):
            raise make_incorrect_min_error(self, self.props.value, value)

        return self.__class__(self.props.update(min=value))

    def max(self, /, value: float) -> "FloatSchema":
        if not isinstance(value, float):
            raise make_invalid_type_error(self, value, (float,))

        if self.props.max is not Nil:
            raise make_already_declared_error(self)

        if (self.props.value is not Nil) and (value < self.props.value):
            raise make_incorrect_max_error(self, self.props.value, value)

        return self.__class__(self.props.update(max=value))



from baby_steps import given, then, when
from pytest import raises

from district42 import optional, schema


def test_dict_get_item():
    with given:
        id_sch = schema.int
        sch = schema.dict({"id": id_sch})

    with when:
        res = sch["id"]

    with then:
        assert res == id_sch


def test_dict_get_optinal_item():
    with given:
        id_sch = schema.int
        sch = schema.dict({optional("id"): id_sch})

    with when:
        res = sch["id"]

    with then:
        assert res == id_sch


def test_dict_get_nested_item():
    with given:
        id_sch = schema.int
        sch = schema.dict({
            "result": schema.dict({
                "id": id_sch
            })
        })

    with when:
        res = sch["result"]["id"]

    with then:
        assert res == id_sch


def test_dict_no_keys_get_nonexisting_item_error():
    with given:
        sch = schema.dict

    with when, raises(Exception) as exception:
        sch["id"]

    with then:
        assert exception.type is KeyError
        assert str(exception.value) == "'id'"


def test_dict_get_nonexisting_item_error():
    with given:
        sch = schema.dict({})

    with when, raises(Exception) as exception:
        sch["id"]

    with then:
        assert exception.type is KeyError
        assert str(exception.value) == "'id'"


def test_dict_get_relaxed_item_error():
    with given:
        sch = schema.dict({...: ...})

    with when, raises(Exception) as exception:
        sch[...]

    with then:
        assert exception.type is KeyError
        assert str(exception.value) == "'...'"



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    from .types import (
        AnySchema,
        BoolSchema,
        ConstSchema,
        DictSchema,
        FloatSchema,
        IntSchema,
        ListSchema,
        NoneSchema,
        StrSchema,
    )

__all__ = ("SchemaVisitor", "SchemaVisitorReturnType",)

SchemaVisitorReturnType = TypeVar("SchemaVisitorReturnType")


class SchemaVisitor(ABC, Generic[SchemaVisitorReturnType]):
    @abstractmethod
    def visit_none(self, schema: "NoneSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_bool(self, schema: "BoolSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_int(self, schema: "IntSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_float(self, schema: "FloatSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_str(self, schema: "StrSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_list(self, schema: "ListSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_dict(self, schema: "DictSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_any(self, schema: "AnySchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    @abstractmethod
    def visit_const(self, schema: "ConstSchema", **kwargs: Any) -> SchemaVisitorReturnType:
        pass

    def __getattr__(self, name: Any) -> Any:
        raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if kwargs.get("extend", False) is not True:
            return
        parent = cls.__bases__[0]
        assert issubclass(parent, SchemaVisitor)
        for name, value in cls.__dict__.items():
            if callable(value) and not name.startswith("__"):
                setattr(parent, name, value)



import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError


def test_str_contains_declaration():
    with given:
        substr = "banana"

    with when:
        sch = schema.str.contains(substr)

    with then:
        assert sch.props.substr == substr


def test_str_invalid_substr_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.contains(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'str', "
                                        "instance of 'int' 42 given")


@pytest.mark.parametrize("substr", [
    "banana",
    "anan",
    "",
])
def test_str_contains_with_value_declaration(substr: str):
    with when:
        sch = schema.str("banana").contains(substr)

    with then:
        assert sch.props.substr == substr


def test_str_contains_already_declared_value_declaration_error():
    with given:
        sch = schema.str("banana!")

    with when, raises(Exception) as exception:
        sch.contains("yellow")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == f"`{sch!r}` does not contain 'yellow'"


def test_str_contains_already_declared_substr_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.contains("banana").contains("banana")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.contains('banana')` is already declared"


def test_str_value_already_declared_contains_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.contains("banana")("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.contains('banana')` is already declared"



from typing import Any

import pytest
from baby_steps import given, then, when

from district42 import represent, schema


def test_const_representation():
    with given:
        sch = schema.const

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.const"


@pytest.mark.parametrize(
    ("value", "expected_repr"),
    [
        (None, "schema.const(None)"),
        (True, "schema.const(True)"),
        (42, "schema.const(42)"),
        (3.14, "schema.const(3.14)"),
        ("banana", "schema.const('banana')")
    ]
)
def test_const_value_representation(value: Any, expected_repr: str):
    with given:
        sch = schema.const(value)

    with when:
        res = represent(sch)

    with then:
        assert res == expected_repr



import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError


@pytest.mark.parametrize("length", [3, 4])
def test_list_contains_elements_len_declaration(length: int):
    with given:
        elements = [schema.int(1), schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements, ...]).len(length)

    with then:
        assert sch.props.elements == [..., *elements, ...]
        assert sch.props.len == length


def test_list_contains_elements_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 1 given"
        )


@pytest.mark.parametrize("min_length", [3, 2])
def test_list_contains_elements_min_len_declaration(min_length: int):
    with given:
        elements = [schema.int(1), schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements, ...]).len(min_length, ...)

    with then:
        assert sch.props.elements == [..., *elements, ...]
        assert sch.props.min_len == min_length


def test_list_contains_elements_min_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(3, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 3 given"
        )


@pytest.mark.parametrize("max_length", [3, 4])
def test_list_contains_elements_max_len_declaration(max_length: int):
    with given:
        elements = [schema.int(1), schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements, ...]).len(..., max_length)

    with then:
        assert sch.props.elements == [..., *elements, ...]
        assert sch.props.max_len == max_length


def test_list_contains_elements_max_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(..., 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max len must be greater than or equal to 2, 1 given"
        )


def test_list_contains_elements_min_max_len_declaration():
    with given:
        elements = [schema.int(1), schema.int(2), schema.int(3)]
        min_length, max_length = 3, 10

    with when:
        sch = schema.list([..., *elements, ...]).len(min_length, max_length)

    with then:
        assert sch.props.elements == [..., *elements, ...]
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


@pytest.mark.parametrize("length", [2, 3])
def test_list_contains_head_elements_len_declaration(length: int):
    with given:
        elements = [schema.int(1), schema.int(2)]

    with when:
        sch = schema.list([*elements, ...]).len(length)

    with then:
        assert sch.props.elements == [*elements, ...]
        assert sch.props.len == length


def test_list_contains_head_elements_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 1 given"
        )


@pytest.mark.parametrize("min_length", [2, 1])
def test_list_contains_head_elements_min_len_declaration(min_length: int):
    with given:
        elements = [schema.int(1), schema.int(2)]

    with when:
        sch = schema.list([*elements, ...]).len(min_length, ...)

    with then:
        assert sch.props.elements == [*elements, ...]
        assert sch.props.min_len == min_length


def test_list_contains_head_elements_min_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(3, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 3 given"
        )


@pytest.mark.parametrize("max_length", [2, 3])
def test_list_contains_head_elements_max_len_declaration(max_length: int):
    with given:
        elements = [schema.int(1), schema.int(2)]

    with when:
        sch = schema.list([*elements, ...]).len(..., max_length)

    with then:
        assert sch.props.elements == [*elements, ...]
        assert sch.props.max_len == max_length


def test_list_contains_head_elements_max_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when, raises(Exception) as exception:
        sch.len(..., 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max len must be greater than or equal to 2, 1 given"
        )


def test_list_contains_head_elements_min_max_len_declaration():
    with given:
        elements = [schema.int(1), schema.int(2)]
        min_length, max_length = 2, 10

    with when:
        sch = schema.list([*elements, ...]).len(min_length, max_length)

    with then:
        assert sch.props.elements == [*elements, ...]
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


@pytest.mark.parametrize("length", [2, 3])
def test_list_contains_tail_elements_len_declaration(length: int):
    with given:
        elements = [schema.int(2), schema.int(3)]
        length = 2

    with when:
        sch = schema.list([..., *elements]).len(length)

    with then:
        assert sch.props.elements == [..., *elements]
        assert sch.props.len == length


def test_list_contains_tail_elements_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 1 given"
        )


@pytest.mark.parametrize("min_length", [2, 1])
def test_list_contains_tail_elements_min_len_declaration(min_length: int):
    with given:
        elements = [schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements]).len(min_length, ...)

    with then:
        assert sch.props.elements == [..., *elements]
        assert sch.props.min_len == min_length


def test_list_contains_tail_elements_min_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(3, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 3 given"
        )


@pytest.mark.parametrize("max_length", [2, 3])
def test_list_contains_tail_elements_max_len_declaration(max_length: int):
    with given:
        elements = [schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements]).len(..., max_length)

    with then:
        assert sch.props.elements == [..., *elements]
        assert sch.props.max_len == max_length


def test_list_contains_tail_elements_max_len_declaration_error():
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(..., 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max len must be greater than or equal to 2, 1 given"
        )


def test_list_contains_tail_elements_min_max_len_declaration():
    with given:
        elements = [schema.int(2), schema.int(3)]
        min_length, max_length = 2, 10

    with when:
        sch = schema.list([..., *elements]).len(min_length, max_length)

    with then:
        assert sch.props.elements == [..., *elements]



import pytest
from baby_steps import given, then, when

from district42 import represent, schema


def test_float_representation():
    with given:
        sch = schema.float

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.float"


@pytest.mark.parametrize(
    ("value", "expected_repr"),
    [
        (3.14, "schema.float(3.14)"),
        (0.0, "schema.float(0.0)"),
        (-3.14, "schema.float(-3.14)"),
    ]
)
def test_float_value_representation(value: float, expected_repr: str):
    with given:
        sch = schema.float(value)

    with when:
        res = represent(sch)

    with then:
        assert res == expected_repr


def test_float_min_value_representation():
    with given:
        sch = schema.float.min(3.14)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.float.min(3.14)"


def test_float_max_value_representation():
    with given:
        sch = schema.float.max(3.14)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.float.max(3.14)"


def test_float_min_max_value_representation():
    with given:
        sch = schema.float.min(3.14).max(3.15)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.float.min(3.14).max(3.15)"


def test_float_min_max_with_value_representation():
    with given:
        sch = schema.float(3.14).min(3.13).max(3.15)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.float(3.14).min(3.13).max(3.15)"



from typing import KeysView

from baby_steps import given, then, when

from district42 import optional, schema


def test_dict_empty_keys():
    with given:
        sch = schema.dict

    with when:
        res = sch.keys()

    with then:
        assert res == KeysView([])


def test_dict_keys():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str,
            optional("email"): schema.str,
        })

    with when:
        res = sch.keys()

    with then:
        assert res == KeysView(["id", "name", "email"])



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import IntSchema


def test_int_declaration():
    with when:
        sch = schema.int

    with then:
        assert isinstance(sch, IntSchema)


def test_int_value_declaration():
    with given:
        value = 42

    with when:
        sch = schema.int(value)

    with then:
        assert sch.props.value == value


def test_int_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.int(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.int` value must be an instance of 'int', "
                                        "instance of 'float' 3.14 given")


def test_int_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.int(42)(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.int(42)` is already declared"


def test_int_min_value_declaration():
    with given:
        min_value = 1

    with when:
        sch = schema.int.min(min_value)

    with then:
        assert sch.props.min == min_value


def test_int_invalid_min_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.min(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.int` value must be an instance of 'int', "
                                        "instance of 'float' 3.14 given")


def test_int_value_already_declared_min_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.min(1)(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.int.min(1)` is already declared"


def test_int_min_value_already_declared_less_value_declaration_error():
    with given:
        sch = schema.int(42)

    with when, raises(Exception) as exception:
        sch.min(43)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min value must be less than or equal to 42, 43 given"
        )


def test_int_min_value_already_declared_min_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.min(1).min(2)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.int.min(1)` is already declared"


def test_int_max_value_declaration():
    with given:
        max_value = 2

    with when:
        sch = schema.int.max(max_value)

    with then:
        assert sch.props.max == max_value


def test_int_invalid_max_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.max(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.int` value must be an instance of 'int', "
                                        "instance of 'float' 3.14 given")


def test_int_value_already_declared_max_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.max(100)(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.int.max(100)` is already declared"


def test_int_max_value_already_declared_greater_value_declaration_error():
    with given:
        sch = schema.int(42)

    with when, raises(Exception) as exception:
        sch.max(41)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max value must be greater than or equal to 42, 41 given"
        )


def test_int_max_value_already_declared_max_declaration_error():
    with when, raises(Exception) as exception:
        schema.int.max(2).max(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.int.max(2)` is already declared"


def test_int_min_max_value_declaration():
    with given:
        min_value, max_value = 1, 2

    with when:
        sch = schema.int.min(min_value).max(max_value)

    with then:
        assert sch.props.min == min_value
        assert sch.props.max == max_value


def test_int_min_max_with_value_declaration():
    with given:
        value = 2
        min_value, max_value = 1, 3

    with when:
        sch = schema.int(value).min(min_value).max(max_value)

    with then:
        assert sch.props.value == value
        assert sch.props.min == min_value
        assert sch.props.max == max_value



from typing import TYPE_CHECKING, Any, Sized, Tuple, Type

if TYPE_CHECKING:
    from ..types import GenericSchema

__all__ = ("DeclarationError", "make_invalid_type_error", "make_already_declared_error",
           "make_incorrect_len_error", "make_incorrect_min_error", "make_incorrect_max_error",
           "make_incorrect_min_len_error", "make_incorrect_max_len_error",)


class DeclarationError(Exception):
    pass


def _get_type_name(type_: Type[Any]) -> str:
    return str(getattr(type_, "__name__", type_))


def make_invalid_type_error(schema: "GenericSchema", value: Any,
                            types: Tuple[Type[Any], ...]) -> DeclarationError:
    value_type = type(value)
    value_type_name = _get_type_name(value_type)

    types_ = tuple(map(_get_type_name, types))
    if len(types) == 1:
        message = (f"`{schema!r}` value must be an instance of '{_get_type_name(types[0])}', "
                   f"instance of {value_type_name!r} {value!r} given")
    else:
        message = (f"`{schema!r}` value must be an instance of {types_!r}, "
                   f"instance of {value_type_name!r} given")
    return DeclarationError(message)


def make_already_declared_error(schema: "GenericSchema") -> DeclarationError:
    message = f"`{schema!r}` is already declared"
    return DeclarationError(message)


def make_incorrect_min_error(schema: "GenericSchema",
                             value: Any, min_value: Any) -> DeclarationError:
    message = f"`{schema!r}` min value must be less than or equal to {value}, {min_value} given"
    return DeclarationError(message)


def make_incorrect_max_error(schema: "GenericSchema",
                             value: Any, max_value: Any) -> DeclarationError:
    message = f"`{schema!r}` max value must be greater than or equal to {value}, {max_value} given"
    return DeclarationError(message)


def make_incorrect_len_error(schema: "GenericSchema",
                             value: Sized, length: int) -> DeclarationError:
    message = f"`{schema!r}` len must be equal to {len(value)}, {length} given"
    return DeclarationError(message)


def make_incorrect_min_len_error(schema: "GenericSchema",
                                 value: Sized, min_length: int) -> DeclarationError:
    message = (f"`{schema!r}` min len must be less than or equal to {len(value)}, "
               f"{min_length} given")
    return DeclarationError(message)


def make_incorrect_max_len_error(schema: "GenericSchema",
                                 value: Sized, max_length: int) -> DeclarationError:
    message = (f"`{schema!r}` max len must be greater than or equal to {len(value)}, "
               f"{max_length} given")
    return DeclarationError(message)



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import BoolSchema


def test_bool_declaration():
    with when:
        sch = schema.bool

    with then:
        assert isinstance(sch, BoolSchema)


def test_bool_value_declaration():
    with given:
        value = True

    with when:
        sch = schema.bool(value)

    with then:
        assert sch.props.value == value


def test_bool_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.bool(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.bool` value must be an instance of 'bool', "
                                        "instance of 'int' 1 given")


def test_bool_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.bool(True)(True)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.bool(True)` is already declared"



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import ListSchema


def test_list_declaration():
    with when:
        sch = schema.list

    with then:
        assert isinstance(sch, ListSchema)


def test_list_empty_elements_declaration():
    with given:
        elements = []

    with when:
        sch = schema.list(elements)

    with then:
        assert sch.props.elements == elements


def test_list_elements_declaration():
    with given:
        elements = [schema.int(1), schema.int(2)]

    with when:
        sch = schema.list(elements)

    with then:
        assert sch.props.elements == elements


def test_list_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list({})

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            "`schema.list` value must be an instance of ('list', 'Schema'), "
            "instance of 'dict' given"
        )


def test_list_invalid_element_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list(["banana"])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            "`schema.list` value must be an instance of ('Schema', 'ellipsis'), "
            "instance of 'str' given"
        )


def test_list_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.list([])([])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list([])` is already declared"


def test_list_contains_any_elements_declaration():
    with when:
        sch = schema.list([...])

    with then:
        assert sch.props.elements == [...]


def test_list_contains_body_elements_declaration():
    with given:
        elements = [schema.int(1), schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements, ...])

    with then:
        assert sch.props.elements == [..., *elements, ...]


def test_list_contains_elements_body_twice_declaration():
    with when, raises(Exception) as exception:
        schema.list([..., ...])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`...` must be first or last element"


def test_list_contains_elements_middle_declaration_error():
    with when, raises(Exception) as exception:
        schema.list([schema.int(1), ..., schema.int(3)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`...` must be first or last element"


def test_list_contains_head_elements_declaration():
    with given:
        elements = [schema.int(1), schema.int(2)]

    with when:
        sch = schema.list([*elements, ...])

    with then:
        assert sch.props.elements == [*elements, ...]


def test_list_contains_elements_head_twice_declaration_error():
    with when, raises(Exception) as exception:
        schema.list([..., ..., schema.int(3)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`...` must be first or last element"


def test_list_contains_tail_elements_declaration():
    with given:
        elements = [schema.int(2), schema.int(3)]

    with when:
        sch = schema.list([..., *elements])

    with then:
        assert sch.props.elements == [..., *elements]


def test_list_contains_elements_tail_twice_declaration_error():
    with when, raises(Exception) as exception:
        schema.list([schema.int(1), ..., ...])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`...` must be first or last element"



from typing import Any, Tuple

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import make_already_declared_error, make_invalid_type_error
from ._schema import GenericSchema, Schema

__all__ = ("AnySchema", "AnyProps",)


class AnyProps(Props):
    @property
    def types(self) -> Nilable[Tuple[GenericSchema, ...]]:
        return self.get("types")


class AnySchema(Schema[AnyProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_any(self, **kwargs)

    def __call__(self, /, type_: GenericSchema, *types: GenericSchema) -> "AnySchema":
        types_ = (type_,) + types
        for t in types_:
            if not isinstance(t, Schema):
                raise make_invalid_type_error(self, t, (Schema,))

        if self.props.types is not Nil:
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(types=types_))



from unittest.mock import sentinel

from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import ListSchema


def test_list_of_elements_declaration():
    with when:
        list_type = schema.int
        sch = schema.list(list_type)

    with then:
        assert isinstance(sch, ListSchema)
        assert sch.props.type == list_type


def test_list_of_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list(sentinel)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            "`schema.list` value must be an instance of ('list', 'Schema'), "
            "instance of '_Sentinel' given"
        )


def test_list_of_len_declaration():
    with given:
        list_type = schema.int
        length = 10

    with when:
        sch = schema.list(list_type).len(length)

    with then:
        assert sch.props.type == list_type
        assert sch.props.len == length


def test_list_of_min_len_declaration():
    with given:
        list_type = schema.int
        min_length = 10

    with when:
        sch = schema.list(list_type).len(min_length, ...)

    with then:
        assert sch.props.type == list_type
        assert sch.props.min_len == min_length


def test_list_of_max_len_declaration():
    with given:
        list_type = schema.int
        max_length = 10

    with when:
        sch = schema.list(list_type).len(..., max_length)

    with then:
        assert sch.props.type == list_type
        assert sch.props.max_len == max_length


def test_list_of_min_max_len_declaration():
    with given:
        list_type = schema.int
        min_length, max_length = 1, 10

    with when:
        sch = schema.list(list_type).len(min_length, max_length)

    with then:
        assert sch.props.type == list_type
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


def test_list_of_value_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1)(schema.str)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1)` is already declared"


def test_list_of_value_already_declared_min_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, ...)(schema.str)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, ...)` is already declared"


def test_list_of_value_already_declared_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(..., 1)(schema.str)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(..., 1)` is already declared"


def test_list_of_value_already_declared_min_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, 2)(schema.str)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, 2)` is already declared"



from baby_steps import given, then, when
from pytest import raises

from district42 import optional


def test_optional():
    with given:
        key = "<key>"
        opt = optional(key)

    with when:
        res = opt.key

    with then:
        assert res == key


def test_optional_error():
    with when, raises(Exception) as exception:
        optional(...)

    with then:
        assert exception.type is TypeError


def test_optional_repr():
    with given:
        opt = optional("<key>")

    with when:
        res = repr(opt)

    with then:
        assert res == "optional('<key>')"



from baby_steps import given, then, when

from district42 import represent, schema


def test_list_representation():
    with given:
        sch = schema.list

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list"


def test_list_empty_representation():
    with given:
        sch = schema.list([])

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list([])"


def test_list_one_element_representation():
    with given:
        sch = schema.list([
            schema.int
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    schema.int",
            "])"
        ])


def test_list_many_elements_representation():
    with given:
        sch = schema.list([
            schema.bool,
            schema.int(1),
            schema.str("banana")
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    schema.bool,",
            "    schema.int(1),",
            "    schema.str('banana')",
            "])"
        ])


def test_list_nested_elements_representation():
    with given:
        sch = schema.list([
            schema.int,
            schema.list([
                schema.int(1),
                schema.int(2),
            ]),
            schema.int
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    schema.int,",
            "    schema.list([",
            "        schema.int(1),",
            "        schema.int(2)",
            "    ]),",
            "    schema.int",
            "])"
        ])


def test_list_contains_no_elements_representation():
    with given:
        sch = schema.list([
            ...
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    ...",
            "])",
        ])


def test_list_contains_body_elements_representation():
    with given:
        sch = schema.list([
            ...,
            schema.int(1),
            schema.int(2),
            ...
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    ...,",
            "    schema.int(1),",
            "    schema.int(2),",
            "    ...",
            "])",
        ])


def test_list_contains_head_elements_representation():
    with given:
        sch = schema.list([
            schema.int(1),
            schema.int(2),
            ...
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    schema.int(1),",
            "    schema.int(2),",
            "    ...",
            "])"
        ])


def test_list_contains_tail_elements_representation():
    with given:
        sch = schema.list([
            ...,
            schema.int(1),
            schema.int(2)
        ])

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "schema.list([",
            "    ...,",
            "    schema.int(1),",
            "    schema.int(2)",
            "])"
        ])


def test_list_len_representation():
    with given:
        sch = schema.list.len(10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list.len(10)"


def test_list_min_len_representation():
    with given:
        sch = schema.list.len(1, ...)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list.len(1, ...)"


def test_list_max_len_representation():
    with given:
        sch = schema.list.len(..., 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list.len(..., 10)"


def test_list_min_max_len_representation():
    with given:
        sch = schema.list.len(1, 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list.len(1, 10)"



import string

import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import StrSchema


def test_str_regex_declaration():
    with given:
        pattern = ".*"

    with when:
        sch = schema.str.regex(pattern)

    with then:
        assert isinstance(sch, StrSchema)
        assert sch.props.pattern == pattern


def test_str_invalid_regex_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.regex(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'str', "
                                        "instance of 'int' 42 given")


def test_str_invalid_regex_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.regex("*")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "Invalid pattern (nothing to repeat at position 0)"


@pytest.mark.parametrize("pattern", [".*", "banana"])
def test_str_pattern_with_value_declaration(pattern: str):

    with when:
        sch = schema.str("banana").regex(pattern)

    with then:
        assert sch.props.pattern == pattern


def test_str_pattern_already_declared_value_declaration_error():
    with given:
        sch = schema.str("banana")
        pattern = "[0-9]+"

    with when, raises(Exception) as exception:
        sch.regex(pattern)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == f"`{sch!r}` does not match {pattern!r}"


@pytest.mark.parametrize("declare", [
    lambda sch: sch.regex(".*"),
    lambda sch: sch("banana"),
    lambda sch: sch.len(1),
    lambda sch: sch.alphabet(string.ascii_letters),
    lambda sch: sch.contains("banana"),
])
def test_str_already_declared_regex_declaration_error(declare):
    with given:
        sch = schema.str.regex(".*")

    with when, raises(Exception) as exception:
        declare(sch)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.regex('.*')` is already declared"



from baby_steps import then, when
from pytest import raises

from district42 import schema
from district42.types import NoneSchema


def test_none_declaration():
    with when:
        sch = schema.none

    with then:
        assert isinstance(sch, NoneSchema)


def test_none_value_declaration_error():
    with when, raises(Exception) as exception:
        schema.none(None)

    with then:
        assert exception.type is TypeError
        assert str(exception.value) == "'NoneSchema' object is not callable"



from baby_steps import given, then, when

from district42 import represent, schema


def test_none_representation():
    with given:
        sch = schema.none

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.none"



from baby_steps import given, then, when

from district42 import represent, schema


def test_any_representation():
    with given:
        sch = schema.any

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.any"


def test_any_type_representation():
    with given:
        sch = schema.any(schema.none)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.any(schema.none)"


def test_any_types_representation():
    with given:
        sch = schema.any(schema.int, schema.str)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.any(schema.int, schema.str)"


def test_any_types_with_values_representation():
    with given:
        sch = schema.any(schema.int(42), schema.str("banana"))

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.any(schema.int(42), schema.str('banana'))"



from baby_steps import given, then, when

from district42 import optional, schema


def test_dict_add():
    with given:
        sch1 = schema.dict({"id": schema.int})
        sch2 = schema.dict({"name": schema.str})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.int,
            "name": schema.str
        })
        assert sch1 == schema.dict({"id": schema.int})
        assert sch2 == schema.dict({"name": schema.str})


def test_dict_add_overide():
    with given:
        sch1 = schema.dict({"id": schema.int})
        sch2 = schema.dict({"id": schema.str})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.str
        })
        assert sch1 == schema.dict({"id": schema.int})
        assert sch2 == schema.dict({"id": schema.str})


def test_dict_add_optional():
    with given:
        sch1 = schema.dict({"id": schema.int})
        sch2 = schema.dict({"name": schema.str, optional("created_at"): schema.int})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.int,
            "name": schema.str,
            optional("created_at"): schema.int
        })
        assert sch1 == schema.dict({"id": schema.int})
        assert sch2 == schema.dict({"name": schema.str, optional("created_at"): schema.int})


def test_dict_add_optional_override():
    with given:
        sch1 = schema.dict({"id": schema.int, optional("created_at"): schema.int})
        sch2 = schema.dict({"name": schema.str, "created_at": schema.int})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.int,
            "name": schema.str,
            "created_at": schema.int
        })
        assert sch1 == schema.dict({"id": schema.int, optional("created_at"): schema.int})
        assert sch2 == schema.dict({"name": schema.str, "created_at": schema.int})


def test_dict_add_relaxed_left():
    with given:
        sch1 = schema.dict({"id": schema.int, ...: ...})
        sch2 = schema.dict({"name": schema.str})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.int,
            ...: ...,
            "name": schema.str
        })
        assert sch1 == schema.dict({"id": schema.int, ...: ...})
        assert sch2 == schema.dict({"name": schema.str})


def test_dict_add_relaxed_right():
    with given:
        sch1 = schema.dict({"id": schema.int})
        sch2 = schema.dict({"name": schema.str, ...: ...})

    with when:
        res = sch1 + sch2

    with then:
        assert res == schema.dict({
            "id": schema.int,
            "name": schema.str,
            ...: ...,
        })
        assert sch1 == schema.dict({"id": schema.int})
        assert sch2 == schema.dict({"name": schema.str, ...: ...})



import pytest
from baby_steps import given, then, when

from district42 import represent, schema


def test_int_representation():
    with given:
        sch = schema.int

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.int"


@pytest.mark.parametrize(
    ("value", "expected_repr"),
    [
        (42, "schema.int(42)"),
        (0, "schema.int(0)"),
        (-42, "schema.int(-42)"),
    ]
)
def test_int_value_representation(value: int, expected_repr: str):
    with given:
        sch = schema.int(value)

    with when:
        res = represent(sch)

    with then:
        assert res == expected_repr


def test_int_min_value_representation():
    with given:
        sch = schema.int.min(42)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.int.min(42)"


def test_int_max_value_representation():
    with given:
        sch = schema.int.max(42)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.int.max(42)"


def test_int_min_max_value_representation():
    with given:
        sch = schema.int.min(1).max(2)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.int.min(1).max(2)"


def test_int_min_max_with_value_representation():
    with given:
        sch = schema.int(2).min(1).max(3)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.int(2).min(1).max(3)"



from typing import Any

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ._schema import Schema

__all__ = ("NoneSchema", "NoneProps",)


class NoneProps(Props):
    pass


class NoneSchema(Schema[NoneProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_none(self, **kwargs)



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import FloatSchema


def test_float_declaration():
    with when:
        sch = schema.float

    with then:
        assert isinstance(sch, FloatSchema)


def test_float_value_declaration():
    with given:
        value = 3.14

    with when:
        sch = schema.float(value)

    with then:
        assert sch.props.value == value


def test_float_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.float("3.14")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.float` value must be an instance of 'float', "
                                        "instance of 'str' '3.14' given")


def test_float_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.float(3.14)(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.float(3.14)` is already declared"


def test_float_min_value_declaration():
    with given:
        min_value = 3.14

    with when:
        sch = schema.float.min(min_value)

    with then:
        assert sch.props.min == min_value


def test_float_invalid_min_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.min("3.14")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.float` value must be an instance of 'float', "
                                        "instance of 'str' '3.14' given")


def test_float_value_already_declared_min_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.min(3.14)(3.2)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.float.min(3.14)` is already declared"


def test_float_min_value_already_declared_less_value_declaration_error():
    with given:
        sch = schema.float(3.14)

    with when, raises(Exception) as exception:
        sch.min(3.15)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min value must be less than or equal to 3.14, 3.15 given"
        )


def test_float_min_value_already_declared_min_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.min(3.14).min(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.float.min(3.14)` is already declared"


def test_float_max_value_declaration():
    with given:
        max_value = 3.14

    with when:
        sch = schema.float.max(max_value)

    with then:
        assert sch.props.max == max_value


def test_float_invalid_max_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.max("3.14")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.float` value must be an instance of 'float', "
                                        "instance of 'str' '3.14' given")


def test_float_value_already_declared_max_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.max(3.2)(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.float.max(3.2)` is already declared"


def test_float_max_value_already_declared_greater_value_declaration_error():
    with given:
        sch = schema.float(3.14)

    with when, raises(Exception) as exception:
        sch.max(3.13)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max value must be greater than or equal to 3.14, 3.13 given"
        )


def test_float_max_value_already_declared_max_declaration_error():
    with when, raises(Exception) as exception:
        schema.float.max(3.14).max(3.14)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.float.max(3.14)` is already declared"


def test_float_min_max_value_declaration():
    with given:
        min_value, max_value = 3.14, 3.15

    with when:
        sch = schema.float.min(min_value).max(max_value)

    with then:
        assert sch.props.min == min_value
        assert sch.props.max == max_value


def test_float_min_max_with_value_declaration():
    with given:
        value = 3.14
        min_value, max_value = 3.13, 3.15

    with when:
        sch = schema.float(value).min(min_value).max(max_value)

    with then:
        assert sch.props.value == value
        assert sch.props.min == min_value
        assert sch.props.max == max_value



import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError


def test_str_len_declaration():
    with given:
        length = 10

    with when:
        sch = schema.str.len(length)

    with then:
        assert sch.props.len == length


def test_str_len_with_value_declaration():
    with given:
        value = "banana"
        length = 6

    with when:
        sch = schema.str(value).len(length)

    with then:
        assert sch.props.value == value
        assert sch.props.len == length


def test_str_len_with_value_declaration_error():
    with when, raises(Exception) as exception:
        schema.str("banana").len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str('banana')` len must be equal to 6, 7 given"


def test_str_invalid_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_str_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7)` is already declared"


def test_str_len_already_declared_min_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(1, ...).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(1, ...)` is already declared"


def test_str_len_already_declared_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(..., 7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(..., 7)` is already declared"


def test_str_len_already_declared_min_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(1, 7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(1, 7)` is already declared"


def test_str_value_already_declared_min_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7, ...)("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7, ...)` is already declared"


def test_str_value_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7)("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7)` is already declared"


def test_str_min_len_declaration():
    with given:
        min_length = 10

    with when:
        sch = schema.str.len(min_length, ...)

    with then:
        assert sch.props.min_len == min_length


@pytest.mark.parametrize("min_length", [6, 5])
def test_str_min_len_with_value_declaration(min_length):
    with given:
        value = "banana"

    with when:
        sch = schema.str(value).len(min_length, ...)

    with then:
        assert sch.props.value == value
        assert sch.props.min_len == min_length


def test_str_min_len_with_value_declaration_error():
    with given:
        value = "banana"
        min_length = 7

    with when, raises(Exception) as exception:
        schema.str(value).len(min_length, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str('banana')` min len must be less than or "
                                        "equal to 6, 7 given")


def test_str_invalid_min_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(None, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_str_min_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7).len(1, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7)` is already declared"


def test_str_max_len_declaration():
    with given:
        max_length = 10

    with when:
        sch = schema.str.len(..., max_length)

    with then:
        assert sch.props.max_len == max_length


@pytest.mark.parametrize("max_length", [6, 7])
def test_str_max_len_with_value_declaration(max_length: int):
    with given:
        value = "banana"

    with when:
        sch = schema.str(value).len(..., max_length)

    with then:
        assert sch.props.value == value
        assert sch.props.max_len == max_length


def test_str_max_len_with_value_declaration_error():
    with given:
        value = "banana"
        max_length = 5

    with when, raises(Exception) as exception:
        schema.str(value).len(..., max_length)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str('banana')` max len must be greater than or "
                                        "equal to 6, 5 given")


def test_str_invalid_max_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(..., None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_str_max_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7).len(..., 7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7)` is already declared"


def test_str_value_already_declared_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(..., 7)("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(..., 7)` is already declared"


def test_str_min_max_len_declaration():
    with given:
        min_length, max_length = 1, 10

    with when:
        sch = schema.str.len(min_length, max_length)

    with then:
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


def test_str_invalid_min_length_type_with_max_length_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(None, 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_str_invalid_max_length_type_with_min_length_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(1, None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_str_min_max_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(7).len(1, 7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(7)` is already declared"


def test_str_value_already_declared_min_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.len(1, 7)("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.len(1, 7)` is already declared"



import pytest
from baby_steps import given, then, when

from district42 import represent, schema


def test_str_representation():
    with given:
        sch = schema.str

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str"


@pytest.mark.parametrize(
    ("value", "expected_repr"),
    [
        ("", "schema.str('')"),
        (" ", "schema.str(' ')"),
        ("banana", "schema.str('banana')"),
    ]
)
def test_str_value_representation(value: str, expected_repr: str):
    with given:
        sch = schema.str(value)

    with when:
        res = represent(sch)

    with then:
        assert res == expected_repr


def test_str_len_representation():
    with given:
        sch = schema.str.len(10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.len(10)"


def test_str_min_len_representation():
    with given:
        sch = schema.str.len(1, ...)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.len(1, ...)"


def test_str_max_len_representation():
    with given:
        sch = schema.str.len(..., 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.len(..., 10)"


def test_str_min_max_len_representation():
    with given:
        sch = schema.str.len(1, 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.len(1, 10)"


def test_str_min_max_len_with_value_representation():
    with given:
        sch = schema.str("banana").len(1, 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str('banana').len(1, 10)"


def test_str_alphabet_representation():
    with given:
        sch = schema.str.alphabet("1234567890")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.alphabet('1234567890')"


def test_str_alphabet_with_value_representation():
    with given:
        sch = schema.str("banana!").alphabet("abn!")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str('banana!').alphabet('abn!')"


def test_str_contains_representation():
    with given:
        sch = schema.str.contains("banana")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.contains('banana')"


def test_str_contains_with_value_representation():
    with given:
        sch = schema.str("banana").contains("banana")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str('banana').contains('banana')"


def test_str_regex_representation():
    with given:
        sch = schema.str.regex(".*")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str.regex('.*')"


def test_str_regex_with_value_representation():
    with given:
        sch = schema.str("banana").regex(".*")

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.str('banana').regex('.*')"



from typing import TYPE_CHECKING, Any, TypeVar, Union

__all__ = ("is_ellipsis", "EllipsisType", "TypeOrEllipsis",)

if TYPE_CHECKING:
    import builtins
    EllipsisType = builtins.ellipsis
else:
    EllipsisType = Any


def is_ellipsis(value: Any) -> bool:
    return isinstance(value, type(...))


_T = TypeVar("_T")
TypeOrEllipsis = Union[_T, EllipsisType]



from typing import Any, Dict, KeysView, Tuple

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import DeclarationError, make_already_declared_error, make_invalid_type_error
from ..utils import TypeOrEllipsis, is_ellipsis
from ._schema import GenericSchema, Schema

__all__ = ("DictSchema", "DictProps", "optional",)


class optional:
    def __init__(self, key: Any) -> None:
        if is_ellipsis(key):
            raise TypeError(key)
        self._key = key

    @property
    def key(self) -> Any:
        return self._key

    def __repr__(self) -> str:
        return f"optional({self._key!r})"


class DictProps(Props):
    @property
    def keys(self) -> Nilable[Dict[Any, Tuple[GenericSchema, bool]]]:
        return self.get("keys")


class DictSchema(Schema[DictProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_dict(self, **kwargs)

    def __call__(self, /, keys: Dict[Any, TypeOrEllipsis[GenericSchema]]) -> "DictSchema":
        if not isinstance(keys, dict):
            raise make_invalid_type_error(self, keys, (dict,))

        if self.props.keys is not Nil:
            raise make_already_declared_error(self)

        real_keys = {}
        for key, val in keys.items():
            if is_ellipsis(key) or is_ellipsis(val):
                if not is_ellipsis(key):
                    message = f"Inappropriate type of key {key!r} ({type(key)!r})"
                    raise DeclarationError(message)
                if not is_ellipsis(val):
                    message = f"Inappropriate type of value {val!r} ({type(val)!r})"
                    raise DeclarationError(message)
            else:
                if not isinstance(val, Schema):
                    raise make_invalid_type_error(self, val, (Schema,))
            if isinstance(key, optional):
                real_keys[key.key] = (val, True)
            else:
                real_keys[key] = (val, False)

        return self.__class__(self.props.update(keys=real_keys))

    def __getitem__(self, /, key: Any) -> GenericSchema:
        if (self.props.keys is Nil) or (key not in self.props.keys) or (is_ellipsis(key)):
            key_repr = "..." if is_ellipsis(key) else key
            raise KeyError(key_repr)
        return self.props.keys[key][0]

    def __add__(self, /, other: "DictSchema") -> "DictSchema":
        self_keys = self.props.keys if (self.props.keys is not Nil) else {}
        other_keys = other.props.keys if (other.props.keys is not Nil) else {}
        merged_keys = {**self_keys, **other_keys}
        return self.__class__(self.props.update(keys=merged_keys))

    def keys(self) -> KeysView[Any]:
        if self.props.keys is Nil:
            return {}.keys()
        return self.props.keys.keys()



from abc import ABC, abstractmethod
from typing import Any, Generic, cast

from niltype import Nil, Nilable

from .._props import PropsType
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType

__all__ = ("Schema", "GenericSchema",)


class Schema(ABC, Generic[PropsType]):
    def __init__(self, props: Nilable[PropsType] = Nil) -> None:
        props_type = self.__orig_bases__[0].__args__[0]  # type: ignore
        self._props = cast(PropsType, props_type()) if props is Nil else props

    @property
    def props(self) -> PropsType:
        return self._props

    @abstractmethod
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        pass

    @classmethod
    def __override__(cls, method: str, fn: Any) -> None:
        setattr(cls, method, fn)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.props!r})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.props == other.props)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __or__(self, other: Any) -> Any:
        raise AttributeError("Schema has no attribute '__or__'")

    def __invert__(self) -> Any:
        raise AttributeError("Schema has no attribute '__invert__'")

    def __mod__(self, other: Any) -> Any:
        raise AttributeError("Schema has no attribute '__mod__'")

    def __add__(self, other: Any) -> Any:
        raise AttributeError("Schema has no attribute '__add__'")


GenericSchema = Schema[Any]



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import StrSchema


def test_str_declaration():
    with when:
        sch = schema.str

    with then:
        assert isinstance(sch, StrSchema)


def test_str_value_declaration():
    with given:
        value = "banana"

    with when:
        sch = schema.str(value)

    with then:
        assert sch.props.value == value


def test_str_invalid_value_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str(None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'str', "
                                        "instance of 'NoneType' None given")


def test_str_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.str("banana")("banana")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str('banana')` is already declared"



import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError


def test_str_alphabet_declaration():
    with given:
        alphabet = "1234567890"

    with when:
        sch = schema.str.alphabet(alphabet)

    with then:
        assert sch.props.alphabet == alphabet


def test_str_invalid_alphabet_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.alphabet(42)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.str` value must be an instance of 'str', "
                                        "instance of 'int' 42 given")


@pytest.mark.parametrize("alphabet", [
    "abn",
    "abnx",
])
def test_str_alphabet_with_value_declaration(alphabet: str):
    with when:
        sch = schema.str("banana").alphabet(alphabet)

    with then:
        assert sch.props.alphabet == alphabet


def test_str_alphabet_already_declared_value_declaration_error():
    with given:
        sch = schema.str("banana!")

    with when, raises(Exception) as exception:
        sch.alphabet("ab")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == f"`{sch!r}` alphabet is missing letters: '!n'"


def test_str_alphabet_already_declared_alphabet_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.alphabet("abn!").alphabet("abn!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.alphabet('abn!')` is already declared"


def test_str_value_already_declared_alphabet_declaration_error():
    with when, raises(Exception) as exception:
        schema.str.alphabet("abn!")("banana!")

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.str.alphabet('abn!')` is already declared"


def test_str_alphabet_and_len_declaration():
    with given:
        alphabet = "1234567890"
        min_length, max_length = 1, 32

    with when:
        sch = schema.str.alphabet(alphabet).len(min_length, max_length)

    with then:
        assert sch.props.alphabet == alphabet
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


def test_str_alphabet_and_contains_declaration():
    with given:
        alphabet = "abn!"
        substr = "banana"

    with when:
        sch = schema.str.alphabet(alphabet).contains(substr)

    with then:
        assert sch.props.alphabet == alphabet
        assert sch.props.substr == substr



from baby_steps import given, then, when

from district42 import represent, schema


def test_list_of_representation():
    with given:
        sch = schema.list(schema.bool)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.bool)"


def test_list_of_values_representation():
    with given:
        sch = schema.list(schema.int(1))

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.int(1))"


def test_list_of_repr_values_representation():
    with given:
        sch = schema.list(schema.str("banana"))

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.str('banana'))"


def test_list_of_len_representation():
    with given:
        sch = schema.list(schema.int).len(10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.int).len(10)"


def test_list_of_min_len_representation():
    with given:
        sch = schema.list(schema.int).len(1, ...)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.int).len(1, ...)"


def test_list_of_max_len_representation():
    with given:
        sch = schema.list(schema.int).len(..., 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.int).len(..., 10)"


def test_list_of_min_max_len_representation():
    with given:
        sch = schema.list(schema.int).len(1, 10)

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.list(schema.int).len(1, 10)"



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError


def test_list_len_declaration():
    with given:
        length = 10

    with when:
        sch = schema.list.len(length)

    with then:
        assert sch.props.len == length


def test_list_invalid_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.list` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_list_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(7)` is already declared"


def test_list_min_len_declaration():
    with given:
        min_length = 10

    with when:
        sch = schema.list.len(min_length, ...)

    with then:
        assert sch.props.min_len == min_length


def test_list_invalid_min_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(None, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.list` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_list_len_already_declared_min_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, ...).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, ...)` is already declared"


def test_list_min_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(7).len(1, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(7)` is already declared"


def test_list_max_len_declaration():
    with given:
        max_length = 10

    with when:
        sch = schema.list.len(..., max_length)

    with then:
        assert sch.props.max_len == max_length


def test_list_invalid_max_length_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(..., None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.list` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_list_invalid_max_length_type_with_min_length_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.list` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_list_len_already_declared_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(..., 7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(..., 7)` is already declared"


def test_list_max_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(7).len(..., 7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(7)` is already declared"


def test_list_min_max_len_declaration():
    with given:
        min_length, max_length = 1, 10

    with when:
        sch = schema.list.len(min_length, max_length)

    with then:
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


def test_list_min_max_same_len_declaration():
    with given:
        min_length, max_length = 5, 5

    with when:
        sch = schema.list.len(min_length, max_length)

    with then:
        assert sch.props.min_len == min_length
        assert sch.props.max_len == max_length


def test_list_invalid_min_length_type_with_max_length_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(None, 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.list` value must be an instance of 'int', "
                                        "instance of 'NoneType' None given")


def test_list_len_already_declared_min_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, 7).len(7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, 7)` is already declared"


def test_list_min_max_len_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(7).len(1, 7)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(7)` is already declared"


def test_list_value_already_declared_min_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, 2)([schema.int(1)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, 2)` is already declared"


def test_list_with_elements_len_declaration():
    with given:
        length = 1

    with when:
        sch = schema.list([schema.int]).len(length)

    with then:
        assert sch.props.len == length


def test_list_value_already_declared_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1)([schema.int(1)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1)` is already declared"


def test_list_with_elements_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == f"`{sch!r}` len must be equal to 2, 1 given"


def test_list_with_elements_min_len_declaration():
    with given:
        length = 1

    with when:
        sch = schema.list([schema.int]).len(length, ...)

    with then:
        assert sch.props.min_len == length


def test_list_value_already_declared_min_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(1, ...)([schema.int(1)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(1, ...)` is already declared"


def test_list_with_elements_min_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(3, ...)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` min len must be less than or equal to 2, 3 given"
        )


def test_list_with_elements_max_len_declaration():
    with given:
        length = 1

    with when:
        sch = schema.list([schema.int]).len(..., length)

    with then:
        assert sch.props.max_len == length


def test_list_value_already_declared_max_len_declaration_error():
    with when, raises(Exception) as exception:
        schema.list.len(..., 1)([schema.int(1)])

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.list.len(..., 1)` is already declared"


def test_list_with_elements_max_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(..., 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max len must be greater than or equal to 2, 1 given"
        )


def test_list_with_elements_min_max_same_len_declaration():
    with given:
        length = 1

    with when:
        sch = schema.list([schema.int]).len(length, length)

    with then:
        assert sch.props.min_len == length
        assert sch.props.max_len == length


def test_list_with_elements_min_max_len_declaration_error():
    with given:
        sch = schema.list([schema.int(1), schema.int(2)])

    with when, raises(Exception) as exception:
        sch.len(2, 1)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == (
            f"`{sch!r}` max len must be greater than or equal to 2, 1 given"
        )



from ._any_schema import AnyProps, AnySchema
from ._bool_schema import BoolProps, BoolSchema
from ._const_schema import ConstProps, ConstSchema
from ._dict_schema import DictProps, DictSchema, optional
from ._float_schema import FloatProps, FloatSchema
from ._int_schema import IntProps, IntSchema
from ._list_schema import ListProps, ListSchema
from ._none_schema import NoneProps, NoneSchema
from ._schema import GenericSchema, Schema
from ._str_schema import StrProps, StrSchema

__all__ = ("AnyProps", "AnySchema", "BoolProps", "BoolSchema", "ConstProps", "ConstSchema",
           "DictProps", "DictSchema", "FloatProps", "FloatSchema",
           "IntProps", "IntSchema", "ListProps", "ListSchema", "NoneProps", "NoneSchema",
           "StrProps", "StrSchema", "GenericSchema", "Schema", "optional",)



from baby_steps import given, then, when
from pytest import raises

from district42 import optional, schema
from district42.errors import DeclarationError


def test_dict_relaxed_empty_keys_declaration():
    with given:
        keys = {...: ...}

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == {...: (..., False)}


def test_dict_relaxed_keys_declaration():
    with given:
        keys = {
            "id": schema.int(42),
            "name": schema.str("banana"),
            ...: ...
        }

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == {key: (val, False) for key, val in keys.items()}


def test_dict_relaxed_optional_keys_declaration():
    with given:
        keys = {
            "id": schema.int(42),
            "name": schema.str("banana"),
            optional("created_at"): schema.int,
            ...: ...,
        }
        props = {
            "id": (schema.int(42), False),
            "name": (schema.str("banana"), False),
            "created_at": (schema.int, True),
            ...: (..., False),
        }

    with when:
        sch = schema.dict(keys)

    with then:
        assert sch.props.keys == props


def test_dict_relaxed_already_declared_declaration_error():
    with given:
        keys = {"key": schema.str}

    with when, raises(Exception) as exception:
        schema.dict({...: ...})(keys)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.dict({...: ...})` is already declared"


def test_dict_already_declared_relaxed_declaration_error():
    with given:
        keys = {"key": schema.str}

    with when, raises(Exception) as exception:
        schema.dict(keys)({...: ...})

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "\n".join([
            "`schema.dict({",
            "    'key': schema.str",
            "})` is already declared",
        ])


def test_dict_relaxed_invalid_key_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.dict({"key": ...})

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "Inappropriate type of key 'key' (<class 'str'>)"


def test_dict_relaxed_invalid_val_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.dict({...: "val"})

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "Inappropriate type of value 'val' (<class 'str'>)"



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import AnySchema


def test_any_declaration():
    with when:
        sch = schema.any

    with then:
        assert isinstance(sch, AnySchema)


def test_any_without_types_declaration_error():
    with when, raises(Exception) as exception:
        schema.any()

    with then:
        assert exception.type is TypeError
        assert "missing 1 required positional argument: 'type_'" in str(exception.value)


def test_any_type_declaration():
    with given:
        type_ = schema.none

    with when:
        sch = schema.any(type_)

    with then:
        assert sch.props.types == (type_,)


def test_any_non_schema_type_declaration_error():
    with when, raises(Exception) as exception:
        schema.any(None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.any` value must be an instance of 'Schema', "
                                        "instance of 'NoneType' None given")


def test_any_types_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.any(schema.none)(schema.none)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.any(schema.none)` is already declared"


def test_any_types_declaration():
    with given:
        types = (schema.int, schema.str,)

    with when:
        sch = schema.any(*types)

    with then:
        assert sch.props.types == types


def test_any_types_with_values_declaration():
    with given:
        types = (schema.int(42), schema.str("banana"),)

    with when:
        sch = schema.any(*types)

    with then:
        assert sch.props.types == types



import pytest
from baby_steps import given, then, when

from district42 import represent, schema


def test_bool_representation():
    with given:
        sch = schema.bool

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.bool"


@pytest.mark.parametrize(
    ("value", "expected_repr"),
    [
        (True, "schema.bool(True)"),
        (False, "schema.bool(False)"),
    ]
)
def test_bool_value_representation(value: bool, expected_repr: str):
    with given:
        sch = schema.bool(value)

    with when:
        res = represent(sch)

    with then:
        assert res == expected_repr



from baby_steps import given, then, when
from pytest import raises

from district42 import schema
from district42.errors import DeclarationError
from district42.types import ConstSchema


def test_const_declaration():
    with when:
        sch = schema.const

    with then:
        assert isinstance(sch, ConstSchema)


def test_const_value_declaration():
    with given:
        value = "banana"

    with when:
        sch = schema.const(value)

    with then:
        assert sch.props.value == value


def test_const_already_declared_declaration_error():
    with when, raises(Exception) as exception:
        schema.const(None)(None)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == "`schema.const(None)` is already declared"
